#ifndef __INCLUDE_KERNEL_CC
#define __INCLUDE_KERNEL_CC

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

using namespace std;
#include "kernel.h"

CKernel::CKernel(char *kparsfilename){
  int iq,ir,ell,dell;
  ParsInit(kparsfilename);
  // dell is the difference in angular kernelum
  dell=1;
  if(IDENTICAL) dell=2;

  kernel=new double **[ellmax+1];
  P=new double [ellmax+1];
  for(ell=0;ell<=ellmax;ell+=1){
    kernel[ell]=NULL;
    if(ell%dell==0){
      kernel[ell]=new double *[nqmax];
      for(iq=0;iq<nqmax;iq++){
	kernel[ell][iq]=new double[nrmax];
	for(ir=0;ir<nrmax;ir++) kernel[ell][iq][ir]=0.0;
      }
    }
  }
}

CKernel::~CKernel(){
  int ell,iq,dell;
  delete [] P;
  for(ell=0;ell<=ellmax;ell+=dell){
    for(iq=0;iq<nqmax;iq++) delete [] kernel[ell][iq];
    if(kernel[ell]!=NULL) delete [] kernel[ell];
  }
  delete [] kernel;
  printf("kernel object deleted\n");
  
}

int CKernel::GetLMAX(){
  return ellmax;
}
double CKernel::GetDELR(){
  return delr;
}
double CKernel::GetDELQ(){
  return delq;
}
int CKernel::GetNRMAX(){
  return nrmax;
}
int CKernel::GetNQMAX(){
  return nqmax;
}
bool CKernel::GetIDENTICAL(){
  return IDENTICAL;
}

double CKernel::GetValue(int ell,double q,double r){
  int iqlow,iqhigh,irlow,irhigh,iq,ir;
  double wrlow,wrhigh,wqlow,wqhigh,answer,ZERO=1.0E-6;
  double qlow,qhigh,rlow,rhigh,qi,ri;
  
  // Perform simple linear extrapolation
  if((IDENTICAL && pow(-1.0,ell)<0) || q>delq*nqmax || r>delr*nrmax) return 0.0;
  else{
    iq=int(floor(q/delq));
    qi=(0.5+iq)*delq;
    if(q<qi){
      iqlow=iq-1;
      if(iqlow<0) iqlow=0;
      iqhigh=iq;
      qhigh=(0.5+iq)*delq;
      wqhigh=fabs(delq-fabs(qhigh-q))/delq;
      if(wqhigh>1.0) wqhigh=1.0;
      wqlow=1.0-wqhigh;
    }
    else{
      iqlow=iq;
      iqhigh=iq+1;
      if(iqhigh>=nqmax) iqhigh=iqlow;
      qlow=(iq+0.5)*delq;
      wqlow=fabs(delq-fabs(q-qlow))/delq;
      if(wqlow>1.0) wqlow=1.0;
      wqhigh=1.0-wqlow;
    }

    ir=int(floor(r/delr));
    ri=(0.5+ir)*delr;
    if(r<ri){
      irlow=ir-1;
      if(irlow<0) irlow=0;
      irhigh=ir;
      rhigh=(0.5+ir)*delr;
      wrhigh=fabs(delr-fabs(rhigh-r))/delr;
      wrhigh=1.0;
      if(wrhigh>1.0) wrhigh=1.0;
      wrlow=1.0-wrhigh;
    }
    else{
      irlow=ir;
      irhigh=ir+1;
      if(irhigh>=nrmax) irhigh=irlow;
      rlow=(ir+0.5)*delr;
      wrlow=fabs(delr-fabs(r-rlow))/delr;
      if(wrlow>1.0) wrlow=1.0;
      wrhigh=1.0-wrlow;
    }

    if(fabs(wrlow+wrhigh-1.0)>1.0E-8 || fabs(wqlow+wqhigh-1.0)>1.0E-6){
      printf("weights do not add up in GetValue(ell,q,r)\n");
      exit(1);
    }

    answer=wqlow*wrlow*GetValue(ell,iqlow,irlow)
      +wqlow*wrhigh*GetValue(ell,iqlow,irhigh)
      +wqhigh*wrlow,GetValue(ell,iqhigh,irlow)
      +wqhigh*wrhigh*GetValue(ell,iqhigh,irhigh);

    if(ell==0 && r>1.0 && answer<-1.0){
      printf("answer below -1, =%g, q=%g, r=%g, iq=%d, ir=%d, nrmax=%d\n",
	     answer,q,r,iq,ir,nrmax);
      //exit(1);
    }
    return answer;
  }
}
double CKernel::GetValue(int ell,int iq,int ir){
  if(iq>=nqmax||ir>=nrmax || iq<0 || ir<0) return 0.0;
  else{
    /* 
       if(ell==0 && ir>=2&& kernel[ell][iq][ir]<-1.01){
       printf("Kernel Warning: ell=0, iq=%d, ir=%d, kernel=%g\n",iq,ir,
       kernel[ell][iq][ir]);
       //exit(1);
      } */
    return kernel[ell][iq][ir];
  }
}

void CKernel::ParsInit(char *kparsfilename){
  parametermap parameters;
  parameter::ReadParsFromFile(parameters,kparsfilename);

  nqmax=parameter::getI(parameters,"NQMAX",25);
  nrmax=parameter::getI(parameters,"NRMAX",120);
  ellmax=parameter::getI(parameters,"KLMAX",24);
  delq=parameter::getD(parameters,"DELQ",4.0);
  delr=parameter::getD(parameters,"DELR",0.5);
  IDENTICAL=parameter::getB(parameters,"IDENTICAL",0);

  printf("reading from %s\n",kparsfilename);
  printf("  _________ PARAMETERS FOR KERNEL ________\n");
  printf("delq set to %g\n",delq);
  printf("nqmax set to %d\n",nqmax);
  printf("delr set to %g\n",delr);
  printf("nrmax set to %d\n",nrmax);
  printf("KLMAX=%d\n",ellmax);
  printf("IDENTICAL set to %d\n",IDENTICAL);
  printf("__________________________________________\n");
}

void CKernel::Read(char *datadir){
  double q,r,delr0;
  int iq,ir,ell,nrmax0,dell;
  char filename[100];
  FILE *infile;

  dell=1;
  if(IDENTICAL) dell=2;
  for(ell=0;ell<=ellmax;ell+=dell){
    for(iq=0;iq<nqmax;iq++){
      q=(0.5+iq)*delq;
      sprintf(filename,"%s/ell%d_q%07.2f.tmp\0",datadir,ell,q);
      infile=fopen(filename,"r");
      fscanf(infile,"%d %lf",&nrmax0,&delr0);
      if(nrmax0<nrmax || fabs(delr-delr0)>1.0E-8){
	printf("Inconsistent values for nrmax or delr in data files!\n");
	exit(1);
      }
      for(ir=0;ir<nrmax;ir++){
	fscanf(infile,"%lf",&kernel[ell][iq][ir]);
      }
      fclose(infile);
    }
  }
}

void CKernel::Write(char *datadir){
  double q,r;
  int ell,iq,ir,dell;
  char filename[100];
  FILE *outfile;
  char shellcommand[120];
  //sprintf(shellcommand,"#!sh; if [ ! -e  %s ]; then mkdir -p %s; fi; \0",
  //  datadir,datadir);
  sprintf(shellcommand,"mkdir -p %s\0",datadir);

  system(shellcommand);

  dell=1;
  if(IDENTICAL) dell=2;
  for(ell=0;ell<=ellmax;ell+=dell){
    for(iq=0;iq<nqmax;iq++){
      q=(0.5+iq)*delq;
      sprintf(filename,"%s/ell%d_q%07.2f.tmp\0",datadir,ell,q);
      printf("For q=%g, Will write to %s\n",q,filename);
      outfile=fopen(filename,"w");
      fprintf(outfile,"%d %g\n",nrmax,delr);
      for(ir=0;ir<nrmax;ir++){
	r=(0.5+ir)*delr;
	fprintf(outfile,"%17.10e\n",kernel[ell][iq][ir]);
      }
      fclose(outfile);
    }
  }
}

void CKernel::Print(){
  double q,r;
  int ell,iq,ir,nrmax0,dell;

  dell=1;
  if(IDENTICAL) dell=2;
  printf("ellmax=%d, nqmax=%d, nrmax=%d\n",ellmax,nqmax,nrmax);
  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    printf("______________ q=%g MeV/c____________________\n",q);
    for(ir=0;ir<nrmax;ir++){
      r=(0.5+ir)*delr;
      printf("%6.2f ",r);
      for(ell=0;ell<=ellmax;ell+=dell){
	printf("%10.3e ",kernel[ell][iq][ir]);
      }
      printf("\n");
    }
  }
}

void CKernel::Calc_ClassCoul(double ma,double mb,int zazb){
  double q,r,ctheta,delctheta;
  int ell,iq,ir,nu=512;
  double x,ea,eb,cweight,u,umax,delu;
      
  for(ell=0;ell<=ellmax;ell++){
    for(iq=0;iq<nqmax;iq++){
      for(ir=0;ir<nrmax;ir++) kernel[ell][iq][ir]=0.0;
    }
  }
  
  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    for(ir=0;ir<nrmax;ir++){
      r=(0.5+ir)*delr;
      ea=sqrt(q*q+ma*ma);
      eb=sqrt(q*q+mb*mb);
      x=2.0*ea*eb*zazb*197.327/(q*q*r*137.036*(ea+eb));

      if(x<1.0){
	umax=2.0*sqrt(1.0-x);
	delu=umax/double(nu);
	for(u=0.5*delu;u<umax;u+=delu){
	  ctheta=-1.0+x+sqrt(u*u+x*x);
	  for(ell=0;ell<=ellmax;ell++)
	    kernel[ell][iq][ir]+=0.5*delu*SpherHarmonics::legendre(ell,ctheta);
	}
	printf("q=%g, r=%g, kernel0=%g =? %g\n",
	       q,r,kernel[0][iq][ir],sqrt(1.0-x));
      }
      kernel[0][iq][ir]-=1.0;
    }
  }
}

void CKernel::Calc_PureHBT(){
  int ir,iq,ell,sign;
  double qr,r,q;

  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    for(ir=0;ir<nrmax;ir++){
      r=(0.5+ir)*delr;
      qr=q*r/197.3269602;
      sign=-1;
      for(ell=0;ell<=ellmax;ell+=2){
	sign=-sign;
	kernel[ell][iq][ir]=2.0*sign*Bessel::jn(ell,qr);
	//printf("ell=%d ellmax=%d, q=%g, r=%g kernel=%g\n",
	//ell,ellmax,q,r,kernel[ell][iq][ir]);
      }
      kernel[0][iq][ir]-=1.0;
    }
  }
}

void CKernel::Calc(CWaveFunction *wf){
  double q,r,ctheta,wf2;
  int ell,dell,iq,ir,ictheta,nctheta=720;

  for(ictheta=0;ictheta<20;ictheta++){
    ctheta=-1.0+2.0*(0.5+ictheta)/20.0;
  }
  dell=1;
  if(IDENTICAL) dell=2; 
  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    for(ir=0;ir<nrmax;ir++){
      r=(0.5+ir)*delr;
      for(ell=0;ell<=ellmax;ell+=dell)
	kernel[ell][iq][ir]=0.0;
      for(ictheta=0;ictheta<nctheta;ictheta++){
	ctheta=-1.0+2.0*(0.5+ictheta)/double(nctheta);
	wf2=wf->GetPsiSquared(q,r,ctheta);
	for(ell=0;ell<=ellmax;ell+=dell){
	  kernel[ell][iq][ir]+=(wf2-1.0)*SpherHarmonics::legendre(ell,ctheta);
	}
      }
      for(ell=0;ell<=ellmax;ell+=dell)
	kernel[ell][iq][ir]=kernel[ell][iq][ir]/double(nctheta);
    }
  }
}

double CKernel::CalcPsiSquared(int iq,int ir,double ctheta){
  int ell,dell=1,nsmall=0;
  double answer=1.0,dela;
  if(IDENTICAL) dell=2;
  ell=0;
  if(ir<nrmax && iq<nqmax){
    while(ell<=ellmax && nsmall<5){
      dela=kernel[ell][iq][ir]*(2.0*ell+1)*P[ell];
      //*SpherHarmonics::legendre(ell,ctheta);
      answer+=dela;
      if(fabs(dela)<1.0E-4) nsmall+=1;
      else(nsmall=0);
      ell+=dell;
    }
  }
  return answer;
}

double CKernel::GetPsiSquared(int iq,int ir,double ctheta){
  int ell,dell=1,nsmall=0;
  double answer=1.0,dela;
  if(ir<nrmax && iq<nqmax){
    CalcP(ctheta);
    answer=CalcPsiSquared(iq,ir,ctheta);
  }
  return answer;
}

double CKernel::CalcPsiSquared(int iq,double r,double ctheta){
  double wa,wb;
  int ira,irb;
  ira=lrint(floor((r/delr)-0.5));
  if(ira<0) ira=0;
  irb=ira+1;
  wb=(r-(ira+0.5)*delr)/delr;
  wa=((irb+0.5)*delr-r)/delr;
  return wa*CalcPsiSquared(iq,ira,ctheta)+wb*CalcPsiSquared(iq,irb,ctheta);
}


double CKernel::GetPsiSquared(int iq,double r,double ctheta){
  double wa,wb,answer=1.0;
  int ira,irb,ell;
  double *P;
  if(iq<nqmax){
    CalcP(ctheta);
    answer=CalcPsiSquared(iq,r,ctheta);
  }
  return answer;
}



double CKernel::GetPsiSquared(double q,double r,double ctheta){
  double wa,wb,answer=1.0;
  int iqa,iqb,ell;
  iqa=lrint(floor((q/delq)-0.5));
  if(iqa<0) iqa=0;
  iqb=iqa+1;
  if(iqb>=nqmax){
    iqb=nqmax-1;
    iqa=iqb-1;
  }
  CalcP(ctheta);
  wb=(q-(iqa+0.5)*delq)/delq;
  wa=((iqb+0.5)*delq-q)/delq;
  answer=wa*CalcPsiSquared(iqa,r,ctheta)+wb*CalcPsiSquared(iqb,r,ctheta);

  return answer;
}

void CKernel::CalcP(double ctheta){
  int ell;
  P[0]=1.0;
  P[1]=ctheta;
  for(ell=1;ell<ellmax;ell++){
    P[ell+1]=((2*ell+1)*ctheta*P[ell]-ell*P[ell-1])/double(ell+1);
  }
}    

// Routines Below for CKernelWF

CKernelWF::CKernelWF(char *kparsfilename){
  int iq,ir,ictheta;
  ParsInit(kparsfilename);
  // dell is the difference in angular kernelum
  delctheta=2.0/double(nctheta);
  if(IDENTICAL) delctheta=delctheta*0.5;
  
  wfarray=new double **[nqmax];
  for(iq=0;iq<nqmax;iq+=1){
    wfarray[iq]=new double *[nrmax];
    for(ir=0;ir<nrmax;ir++){
      wfarray[iq][ir]=new double[nctheta+1];
      for(ictheta=0;ictheta<=nctheta;ictheta++) wfarray[iq][ir][ictheta]=1.0;
    }
  }
}

void CKernelWF::ParsInit(char *kparsfilename){
  parametermap parameters;
  parameter::ReadParsFromFile(parameters,kparsfilename);

  nqmax=parameter::getI(parameters,"NQMAX",25);
  nrmax=parameter::getI(parameters,"NRMAX",120);
  nctheta=parameter::getI(parameters,"NCTHETA",120);
  delq=parameter::getD(parameters,"DELQ",4.0);
  delr=parameter::getD(parameters,"DELR",0.5);
  IDENTICAL=parameter::getB(parameters,"IDENTICAL",0);

  printf("reading from %s\n",kparsfilename);
  printf("  _________ PARAMETERS FOR KERNELWF ________\n");
  printf("delq set to %g\n",delq);
  printf("nqmax set to %d\n",nqmax);
  printf("delr set to %g\n",delr);
  printf("nrmax set to %d\n",nrmax);
  printf("nctheta set to %d\n",nctheta);
  printf("IDENTICAL set to %d\n",IDENTICAL);
  printf("__________________________________________\n");
}


CKernelWF::~CKernelWF(){
  int iq,ir;
  for(iq=0;iq<nqmax;iq++){
    for(ir=0;ir<nrmax;ir++) delete [] wfarray[iq][ir];
    delete [] wfarray[iq];
  }
  delete [] wfarray;
  printf("KernelWF object deleted\n");
  
}

void CKernelWF::Calc(CWaveFunction *wf){
  int iq,ir,ictheta;
  double q,r,ctheta,ps2;

  for(iq=0;iq<nqmax;iq++){
    q=(iq+0.5)*delq;
    for(ir=0;ir<nrmax;ir++){
      r=(ir+0.5)*delr;
      for(ictheta=0;ictheta<=nctheta;ictheta++){
	ctheta=1.0-ictheta*delctheta;
	ps2=wf->GetPsiSquared(q,r,ctheta);
	if(ps2<0.0 && r>1.0){
	  printf("screwy, psi^2=%g, r=%g, q=%g, ctheta=%g\n",ps2,r,q,ctheta);
	  exit(1);
	}
	wfarray[iq][ir][ictheta]=ps2-1.0;
      }
    }
  }
}

double CKernelWF::GetPsiSquared(int iq,int ir,int ictheta){
  if(iq<nqmax && ir<nrmax && ictheta<=nctheta)
    return 1.0+wfarray[iq][ir][ictheta];
  else return 1.0;  
}
double CKernelWF::GetPsiSquared(int iq,int ir,double ctheta){
  double wa,wb;
  int ictheta;
  if(iq<nqmax && ir<nrmax){
    if(IDENTICAL) ctheta=fabs(ctheta);
    ictheta=lrint(floor((1.0-ctheta)/delctheta));
    wb=((1.0-ctheta)-ictheta*delctheta)/delctheta;
    wa=1.0-wb;
    return 1.0+wa*wfarray[iq][ir][ictheta]+wb*wfarray[iq][ir][ictheta+1];
  }
  else return 1.0;
}

double CKernelWF::GetPsiSquared(int iq,double r,double ctheta){
  int ir;
  double wa,wb;
  if(iq<nqmax){
    ir=lrint(floor((r-0.5*delr)/delr));
    if(ir<0) ir=0;
    wb=(r-(ir+0.5)*delr)/delr;
    wa=1.0-wb;
    return 1.0+wa*GetPsiSquared(iq,ir,ctheta)+wb*GetPsiSquared(iq,ir+1,ctheta);
  }
  else return 1.0;
}

double CKernelWF::GetPsiSquared(double q,double r,double ctheta){
  int iq;
  double wa,wb;
  iq=lrint(floor((q-0.5*delq)/delq));
  if(iq<0) iq=0;
  wb=(q-(iq+0.5)*delq)/delq;
  wa=1.0-wb;
  return wa*GetPsiSquared(iq,r,ctheta)+wb*GetPsiSquared(iq+1,r,ctheta);
}

double CKernelWF::GetDELR(){
  return delr;
}
double CKernelWF::GetDELQ(){
  return delq;
}
double CKernelWF::GetDELCTHETA(){
  return delctheta;
}
int CKernelWF::GetNRMAX(){
  return nrmax;
}
int CKernelWF::GetNQMAX(){
  return nqmax;
}
int CKernelWF::GetNCTHETA(){
  return nctheta;
}
bool CKernelWF::GetIDENTICAL(){
  return IDENTICAL;
}

void CKernelWF::Write(char *datadir){
  double q,r;
  int iq,ir,ictheta;
  char filename[80];
  FILE *outfile;
  char shellcommand[120];
  //sprintf(shellcommand,"#!sh; if [ ! -e  %s ]; then mkdir -p %s; fi; \0",
  //  datadir,datadir);
  sprintf(shellcommand,"mkdir -p %s\0",datadir);
  system(shellcommand);
  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    sprintf(filename,"%s/q%07.2f.tmp\0",datadir,q);
    outfile=fopen(filename,"w");
    fprintf(outfile,"%d %d\n",nrmax,nctheta);
    for(ir=0;ir<nrmax;ir++){
      for(ictheta=0;ictheta<=nctheta;ictheta++){
	fprintf(outfile,"%17.10e ",wfarray[iq][ir][ictheta]);
      }
      fprintf(outfile,"\n");
    }
    fclose(outfile);
  } 
}

void CKernelWF::Read(char *datadir){
  double q,r;
  int iq,ir,ictheta,nrmaxread,ncthetaread;
  char filename[80];
  FILE *outfile;
  nrmaxread,ncthetaread;

  for(iq=0;iq<nqmax;iq++){
    q=(0.5+iq)*delq;
    sprintf(filename,"%s/q%07.2f.tmp\0",datadir,q);
    outfile=fopen(filename,"r");
    fscanf(outfile,"%d %d\n",&nrmaxread,&ncthetaread);
    if(nrmaxread!=nrmax || ncthetaread!=nctheta){
      printf("CKernelWF : trying to read file with wrong dimensions\n");
      printf("nrmaxread=%d, nrmax=%d, ncthetaread=%d, nctheta=%d\n",
	     nrmaxread,nrmax,ncthetaread,nctheta);
      exit(1);
    }
    for(ir=0;ir<nrmax;ir++){
      for(ictheta=0;ictheta<=nctheta;ictheta++){
	fscanf(outfile,"%lf ",&wfarray[iq][ir][ictheta]);
      }
    }
    fclose(outfile);
  } 
}

#endif

